from sage.all import EllipticCurve
from sage.all import Integer
from sage.arith.misc import inverse_mod
from sage.rings.finite_rings.integer_mod import *
from sage.misc.functional import lift

from montgomery import *
from theta_isogeny import *
from compute_gamma import *
from utilities.discrete_log import *
from utilities.pairing import *
from utilities.supersingular import *


def compressed_pubkey(pubkey,a,b,c,Alice_secret,p,Fp,Fp2):
    
    Pa = pubkey[0][0]
    Qa = pubkey[0][1]
    Pb = pubkey[0][2]
    Qb = pubkey[0][3]
    PQb = pubkey[0][4]
    R = pubkey[0][5]
    A3 = get_A(Pb,Qb,PQb)

    Pa = scalar(Pa,A3,Alice_secret[0])
    Qa = scalar(Qa,A3,Alice_secret[1])

    S = MontoEllip(A3,[Pa,Qa,Pb,Qb,PQb,R])

    PQcheck = S[1][2] + S[1][3]
    if (PQcheck != S[1][4]) and (PQcheck != (-1)*S[1][4]):
        S[1][3] = (-1)*S[1][3]

    a_ex = 0
    check_ex = (p+1)//(2**(a+2))
    while check_ex % 2 == 0:
        a_ex += 1
        check_ex = check_ex // 2

    b_ex = 0
    check_ex = (p+1)//(3**b)
    while check_ex % 3 == 0:
        b_ex += 1
        check_ex = check_ex // 3

    c_ex = 0
    check_ex = (p+1)//(5**c)
    while check_ex % 5 == 0:
        c_ex += 1
        check_ex = check_ex // 5

    D2 = 2**(a+2+a_ex)
    D3 = 3**(b+b_ex)
    D5 = 5**(c+c_ex)

    P, Q = torsion_basis(S[0], D2*D3*D5)

    #Compression of Pa Qa
    P2 = (D3*D5) * P
    Q2 = (D3*D5) * Q

    window = [(a+2)//10]

    e = (p**2 - 1)//D2

    criteria_a = tate_pairing_pari(P2,Q2,D2)**e
    
    Pae1 = tate_pairing_pari(S[1][0],Q2,D2)**e
    Pae2 = tate_pairing_pari(P2,S[1][0],D2)**e
    Qae1 = tate_pairing_pari(S[1][1],Q2,D2)**e
    Qae2 = tate_pairing_pari(P2,S[1][1],D2)**e

    Pa_v1 = windowed_pohlig_hellman(Pae1,criteria_a,a+2+a_ex,window)//(2**a_ex)
    Pa_v2 = windowed_pohlig_hellman(Pae2,criteria_a,a+2+a_ex,window)//(2**a_ex)
    Qa_v1 = windowed_pohlig_hellman(Qae1,criteria_a,a+2+a_ex,window)//(2**a_ex)
    Qa_v2 = windowed_pohlig_hellman(Qae2,criteria_a,a+2+a_ex,window)//(2**a_ex)

    matrix_A1 = [Qa_v2,(-1)*Pa_v2,(-1)*Qa_v1,Pa_v1]

    point_checker = (2**(a+a_ex)) * P2
    point_checker2 = (2**(a+a_ex)) * Q2
    point_checker3 = 2 * point_checker2

    matrix_A = [Alice_secret[0]*matrix_A1[0],Alice_secret[0]*matrix_A1[1],Alice_secret[1]*matrix_A1[2],Alice_secret[1]*matrix_A1[3]]

    if point_checker[0] == Fp2(-1):
        matrix_A[0] += 2*(Alice_secret[0] - Alice_secret[1])*matrix_A1[2]
        matrix_A[1] += 2*(Alice_secret[0] - Alice_secret[1])*matrix_A1[3]

    elif point_checker3[0] == Fp2(0):
        matrix_A = [Alice_secret[1]*matrix_A1[0],Alice_secret[1]*matrix_A1[1],Alice_secret[0]*matrix_A1[2],Alice_secret[0]*matrix_A1[3]]

        if point_checker2[0] == Fp2(-1):
            matrix_A[2] += 2*(Alice_secret[0] - Alice_secret[1])*matrix_A1[0]
            matrix_A[3] += 2*(Alice_secret[0] - Alice_secret[1])*matrix_A1[1]

    elif point_checker[0] != Fp2(1):
        matrix_A[0] += (Alice_secret[0] - Alice_secret[1])*matrix_A1[2]
        matrix_A[1] += (Alice_secret[0] - Alice_secret[1])*matrix_A1[3]
        point_checker4 = point_checker + point_checker2
        if point_checker4[0] == Fp2(-1):
            matrix_A[0] += 2*(Alice_secret[0] - Alice_secret[1])*matrix_A1[2]
            matrix_A[1] += 2*(Alice_secret[0] - Alice_secret[1])*matrix_A1[3]

    #Compression of Pb Qb
    P3 = (D2*D5) * P
    Q3 = (D2*D5) * Q

    e = (p**2 - 1)//D3

    criteria_b = tate_pairing_pari(P3,Q3,D3)**e
    
    Pbe1 = tate_pairing_pari(S[1][2],Q3,D3)**e
    Pbe2 = tate_pairing_pari(P3,S[1][2],D3)**e
    Qbe1 = tate_pairing_pari(S[1][3],Q3,D3)**e
    Qbe2 = tate_pairing_pari(P3,S[1][3],D3)**e

    Pb_v1 = discrete_log_pari(Pbe1,criteria_b,D3)//(3**b_ex)
    Pb_v2 = discrete_log_pari(Pbe2,criteria_b,D3)//(3**b_ex)
    Qb_v1 = discrete_log_pari(Qbe1,criteria_b,D3)//(3**b_ex)
    Qb_v2 = discrete_log_pari(Qbe2,criteria_b,D3)//(3**b_ex)

    matrix_B = [Qb_v2,(-1)*Pb_v2,(-1)*Qb_v1,Pb_v1]

    #Compression of R
    P5 = (D2*D3) * P
    Q5 = (D2*D3) * Q

    e = (p**2 - 1)//D5

    criteria_c = tate_pairing_pari(P5,Q5,D5)**e
    
    Rce1 = tate_pairing_pari(S[1][5],Q5,D5)**e
    Rce2 = tate_pairing_pari(P5,S[1][5],D5)**e

    Rc_v1 = discrete_log_pari(Rce1,criteria_c,D5)//(5**c_ex)
    Rc_v2 = discrete_log_pari(Rce2,criteria_c,D5)//(5**c_ex)

    if Rc_v1 % 5 == 0:

        inv_R = inverse_mod(Rc_v2,5**c)
        Rc_v1 = (Rc_v1 * inv_R) % 5**c
        Rc = [0,Rc_v1]

    else:

        inv_R = inverse_mod(Rc_v1,5**c)
        Rc_v2 = (Rc_v2 * inv_R) % 5**c
        Rc = [1,Rc_v2]

    #Compression of Pa1 Qa1
    Pa1 = pubkey[1][0]
    Qa1 = pubkey[1][1]
    Pb1 = pubkey[1][2]
    Qb1 = pubkey[1][3]
    PQb1 = pubkey[1][4]
    R1 = pubkey[1][5]
    A4 = get_A(Pb1,Qb1,PQb1)

    S1 = MontoEllip(A4,[Pa1,Qa1,Pb1,Qb1,PQb1,R1])

    if (weil_pairing_pari(S[1][0],S[1][1],2**(a+2)))**(2**(2*a)-9) != weil_pairing_pari(S1[1][0],S1[1][1],2**(a+2)):
        S1[1][0] = (-1)*S1[1][0]

    PQcheck = S1[1][2] + S1[1][3]
    if (PQcheck != S1[1][4]) and (PQcheck != (-1)*S1[1][4]):
        S1[1][3] = (-1)*S1[1][3]

    P, Q = torsion_basis(S1[0], D2*D3*D5)

    P2 = (D3*D5) * P
    Q2 = (D3*D5) * Q

    e = (p**2 - 1)//D2

    criteria_a = tate_pairing_pari(P2,Q2,D2)**e
    
    Pae1 = tate_pairing_pari(S1[1][0],Q2,D2)**e
    Pae2 = tate_pairing_pari(P2,S1[1][0],D2)**e
    Qae1 = tate_pairing_pari(S1[1][1],Q2,D2)**e
    Qae2 = tate_pairing_pari(P2,S1[1][1],D2)**e

    Pa_v1 = windowed_pohlig_hellman(Pae1,criteria_a,a+2+a_ex,window)//(2**a_ex)
    Pa_v2 = windowed_pohlig_hellman(Pae2,criteria_a,a+2+a_ex,window)//(2**a_ex)
    Qa_v1 = windowed_pohlig_hellman(Qae1,criteria_a,a+2+a_ex,window)//(2**a_ex)
    Qa_v2 = windowed_pohlig_hellman(Qae2,criteria_a,a+2+a_ex,window)//(2**a_ex)

    Pa_v1_1 = matrix_A[0]*Pa_v1 + matrix_A[1]*Qa_v1
    Pa_v2_1 = matrix_A[0]*Pa_v2 + matrix_A[1]*Qa_v2
    Qa_v1_1 = matrix_A[2]*Pa_v1 + matrix_A[3]*Qa_v1
    Qa_v2_1 = matrix_A[2]*Pa_v2 + matrix_A[3]*Qa_v2

    if Pa_v1_1 % 2 == 0:
        inv_A = inverse_mod(Pa_v2_1,2**(a+2))
        
        Pa_v1_1 = (Pa_v1_1 * inv_A) % 2**(a+2)
        Qa_v1_1 = (Qa_v1_1 * inv_A) % 2**(a+2)
        Qa_v2_1 = (Qa_v2_1 * inv_A) % 2**(a+2)
        phiA = [0,Pa_v1_1,Qa_v1_1,Qa_v2_1]
        
    else:
        inv_A = inverse_mod(Pa_v1_1,2**(a+2))
        
        Pa_v2_1 = (Pa_v2_1 * inv_A) % 2**(a+2)
        Qa_v1_1 = (Qa_v1_1 * inv_A) % 2**(a+2)
        Qa_v2_1 = (Qa_v2_1 * inv_A) % 2**(a+2)
        phiA = [1,Pa_v2_1,Qa_v1_1,Qa_v2_1] 

    #Compression of Pb1 Qb1
    P3 = (D2*D5) * P
    Q3 = (D2*D5) * Q

    e = (p**2 - 1)//D3
    
    criteria_b = tate_pairing_pari(P3,Q3,D3)** e
    
    Pbe1 = tate_pairing_pari(S1[1][2],Q3,D3)**e
    Pbe2 = tate_pairing_pari(P3,S1[1][2],D3)**e
    Qbe1 = tate_pairing_pari(S1[1][3],Q3,D3)**e
    Qbe2 = tate_pairing_pari(P3,S1[1][3],D3)**e

    Pb_v1 = discrete_log_pari(Pbe1,criteria_b,D3)//(3**b_ex)
    Pb_v2 = discrete_log_pari(Pbe2,criteria_b,D3)//(3**b_ex)
    Qb_v1 = discrete_log_pari(Qbe1,criteria_b,D3)//(3**b_ex)
    Qb_v2 = discrete_log_pari(Qbe2,criteria_b,D3)//(3**b_ex)

    Pb_v1_1 = matrix_B[0]*Pb_v1 + matrix_B[1]*Qb_v1
    Pb_v2_1 = matrix_B[0]*Pb_v2 + matrix_B[1]*Qb_v2
    Qb_v1_1 = matrix_B[2]*Pb_v1 + matrix_B[3]*Qb_v1
    Qb_v2_1 = matrix_B[2]*Pb_v2 + matrix_B[3]*Qb_v2

    if Pb_v1_1 % 3 == 0:

        inv_B = inverse_mod(Pb_v2_1,3**b)
        
        Pb_v1_1 = (Pb_v1_1 * inv_B) % 3**b
        Qb_v1_1 = (Qb_v1_1 * inv_B) % 3**b
        Qb_v2_1 = (Qb_v2_1 * inv_B) % 3**b
        
        phiB = [0,Pb_v1_1,Qb_v1_1,Qb_v2_1]
        
    else:

        inv_B = inverse_mod(Pb_v1_1,3**b)
        
        Pb_v2_1 = (Pb_v2_1 * inv_B) % 3**b
        Qb_v1_1 = (Qb_v1_1 * inv_B) % 3**b
        Qb_v2_1 = (Qb_v2_1 * inv_B) % 3**b
        
        phiB = [1,Pb_v2_1,Qb_v1_1,Qb_v2_1]

    #Compression of Rc1
    P5 = (D2*D3) * P
    Q5 = (D2*D3) * Q

    e = (p**2 - 1)//D5

    criteria_c = tate_pairing_pari(P5,Q5,D5)**e
    
    Rce1 = tate_pairing_pari(S1[1][5],Q5,D5)**e
    Rce2 = tate_pairing_pari(P5,S1[1][5],D5)**e

    Rc_v1 = discrete_log_pari(Rce1,criteria_c,D5)//(5**c_ex)
    Rc_v2 = discrete_log_pari(Rce2,criteria_c,D5)//(5**c_ex)

    Rc_v1_1 = (inv_R * Rc_v1)
    Rc_v2_1 = (inv_R * Rc_v2)

    if Rc_v1_1 % 5 == 0:
        inv_R = inverse_mod(Rc_v2_1,5**c)
        Rc_v1_1 = (Rc_v1_1 * inv_R) % 5**c
        phiRc = [0,Rc_v1_1]

    else:
        inv_R = inverse_mod(Rc_v1_1,5**c)
        Rc_v2_1 = (Rc_v2_1 * inv_R) % 5**c
        phiRc = [1,Rc_v2_1]
    
    Alice_secret[2] = (Alice_secret[2] * inv_R) % 5**c
    
    return [[A3,Rc],[A4,phiA,phiB,phiRc]],Alice_secret


def decomp_pubkey(pubkey_comp,a,b,c,p,Fp,Fp2):

    #For E
    S = MontoEllip(pubkey_comp[0][0],[])

    a_ex = 0
    check_ex = (p+1)//(2**(a+2))
    while check_ex % 2 == 0:
        a_ex += 1
        check_ex = check_ex // 2

    b_ex = 0
    check_ex = (p+1)//(3**b)
    while check_ex % 3 == 0:
        b_ex += 1
        check_ex = check_ex // 3

    c_ex = 0
    check_ex = (p+1)//(5**c)
    while check_ex % 5 == 0:
        c_ex += 1
        check_ex = check_ex // 5

    P, Q = torsion_basis(S[0], (2**(a+2+a_ex)) * (3**(b+b_ex)) * (5**(c+c_ex)))

    D25 = (2**(a+2+a_ex)) * (3**b_ex) * (5**(c+c_ex))
    D23 = (2**(a+2+a_ex)) * (3**(b+b_ex)) * (5**c_ex)
    D35 = (2**a_ex) * (3**(b+b_ex)) * (5**(c+c_ex))

    Pb_point = D25 * P
    Qb_point = D25 * Q
    PQb_point = Pb_point + Qb_point

    Pb = [Pb_point[0],Pb_point[2]]
    Qb = [Qb_point[0],Qb_point[2]]
    PQb = [PQb_point[0],PQb_point[2]]

    if pubkey_comp[0][1][0] == 0:
        Rc_point = D23 * pubkey_comp[0][1][1] * P + D23 * Q
    else:
        Rc_point = D23 * P + D23 * pubkey_comp[0][1][1] * Q
    
    Rc = [Rc_point[0],Rc_point[2]]

    Pa_point = D35 * P
    Qa_point = D35 * Q
    
    point_checker = (2**a) * Pa_point
    point_checker2 = (2**a) * Qa_point
    point_checker3 = 2 * point_checker2

    matrix_A = [1,0,0,1]

    if point_checker[0] == Fp2(-1):
        Pa_point = Pa_point + 2 * Qa_point
        matrix_A = [1,2,0,1]

    elif point_checker3[0] == Fp2(0):
        Pa_point_sub = Pa_point
        Pa_point = Qa_point
        Qa_point = Pa_point_sub
        matrix_A = [0,1,1,0]

        if point_checker2[0] == Fp2(-1):
            Pa_point = Pa_point + 2 * Qa_point
            matrix_A = [2,1,1,0]

    elif point_checker[0] != Fp2(1):
        Pa_point += Qa_point
        point_checker4 = 2**a * Pa_point
        matrix_A = [1,1,0,1]
        if point_checker4[0] == Fp2(-1):
            Pa_point = Pa_point + 2 * Qa_point
            matrix_A = [1,3,0,1]

    Pa = [Pa_point[0],Pa_point[2]]
    Qa = [Qa_point[0],Qa_point[2]]

    #For E1
    S = MontoEllip(pubkey_comp[1][0],[])
    P, Q = torsion_basis(S[0], (2**(a+2+a_ex)) * (3**(b+b_ex)) * (5**(c+c_ex)))

    P1 = D25 * P
    Q1 = D25 * Q

    if pubkey_comp[1][2][0] == 0:
        phiPb_point = (pubkey_comp[1][2][1]) * P1 + Q1
        phiQb_point = (pubkey_comp[1][2][2]) * P1 + (pubkey_comp[1][2][3]) * Q1

        phiPQb_point = phiPb_point + phiQb_point

        phiPb = [phiPb_point[0],phiPb_point[2]]
        phiQb = [phiQb_point[0],phiQb_point[2]]
        phiPQb = [phiPQb_point[0],phiPQb_point[2]]

    else:
        phiPb_point = P1 + (pubkey_comp[1][2][1]) * Q1
        phiQb_point = (pubkey_comp[1][2][2]) * P1 + (pubkey_comp[1][2][3]) * Q1

        phiPQb_point = phiPb_point + phiQb_point

        phiPb = [phiPb_point[0],phiPb_point[2]]
        phiQb = [phiQb_point[0],phiQb_point[2]]
        phiPQb = [phiPQb_point[0],phiPQb_point[2]]

    P5 = D23 * P
    Q5 = D23 * Q

    if pubkey_comp[1][3][0] == 0:
        phiRc_point = pubkey_comp[1][3][1] * P5 + Q5

        phiRc = [phiRc_point[0],phiRc_point[2]]

    else:
        phiRc_point = P5 + (pubkey_comp[1][3][1]) * Q5

        phiRc = [phiRc_point[0],phiRc_point[2]]

    P2 = D35 * P
    Q2 = D35 * Q

    if pubkey_comp[1][1][0] == 0:
        phiPa_point = (pubkey_comp[1][1][1]) * P2 + Q2
        phiQa_point = (pubkey_comp[1][1][2]) * P2 + (pubkey_comp[1][1][3]) * Q2

    else:
        phiPa_point = P2 + (pubkey_comp[1][1][1]) * Q2
        phiQa_point = (pubkey_comp[1][1][2]) * P2 + (pubkey_comp[1][1][3]) * Q2

    phiPa_point1 = matrix_A[0] * phiPa_point + matrix_A[1] * phiQa_point
    phiQa_point1 = matrix_A[2] * phiPa_point + matrix_A[3] * phiQa_point
    
    phiPa = [phiPa_point1[0],phiPa_point1[2]]
    phiQa = [phiQa_point1[0],phiQa_point1[2]]

    return [[Pa,Qa,Pb,Qb,PQb,Rc],[phiPa,phiQa,phiPb,phiQb,phiPQb,phiRc]]


def compressed_cipher(cipher,a,c,p,Fp,Fp2):

    S1 = MontoEllip(cipher[0][0],[cipher[0][1],cipher[0][2],cipher[0][3]])

    a_ex = 0
    check_ex = (p+1)//(2**(a+2))
    while check_ex % 2 == 0:
        a_ex += 1
        check_ex = check_ex // 2

    c_ex = 0
    check_ex = (p+1)//(5**c)
    while check_ex % 5 == 0:
        c_ex += 1
        check_ex = check_ex // 5

    D2 = 2**(a+2+a_ex)
    D5 = 5**(c+c_ex)
    #P, Q = entangled_torsion_basis(S1[0], elligator_tables, (p+1)//(D2*D5))
    P, Q = torsion_basis(S1[0], D2*D5)

    P1 = D5 * P
    Q1 = D5 * Q
    P5 = D2 * P
    Q5 = D2 * Q


    #Compression of P1 Q1
    e = (p**2 - 1)//D2

    criteria_1 = tate_pairing_pari(P1,Q1,D2)** e
    
    phiPae1 = tate_pairing_pari(S1[1][0],Q1,D2)** e
    phiPae2 = tate_pairing_pari(P1,S1[1][0],D2)** e
    phiQae1 = tate_pairing_pari(S1[1][1],Q1,D2)** e
    phiQae2 = tate_pairing_pari(P1,S1[1][1],D2)** e

    window = [(a+2)//10]

    phiPa_v1 = windowed_pohlig_hellman(phiPae1,criteria_1,a+2+a_ex,window)//(2**a_ex)
    phiPa_v2 = windowed_pohlig_hellman(phiPae2,criteria_1,a+2+a_ex,window)//(2**a_ex)
    phiQa_v1 = windowed_pohlig_hellman(phiQae1,criteria_1,a+2+a_ex,window)//(2**a_ex)
    phiQa_v2 = windowed_pohlig_hellman(phiQae2,criteria_1,a+2+a_ex,window)//(2**a_ex)

    if phiPa_v1 % 2 == 0:

        inv_phi = inverse_mod(phiPa_v2,2**(a+2))
        
        phiPa_v1 = (phiPa_v1 * inv_phi) % 2**(a+2)
        phiQa_v1 = (phiQa_v1 * inv_phi) % 2**(a+2)
        phiQa_v2 = (phiQa_v2 * inv_phi) % 2**(a+2)
        
        phiA1 = [0,phiPa_v1,phiQa_v1,phiQa_v2]

    else:

        inv_phi = inverse_mod(phiPa_v1,2**(a+2))
        
        phiPa_v2 = (phiPa_v2 * inv_phi) % 2**(a+2)
        phiQa_v1 = (phiQa_v1 * inv_phi) % 2**(a+2)
        phiQa_v2 = (phiQa_v2 * inv_phi) % 2**(a+2)
        
        phiA1 = [1,phiPa_v2,phiQa_v1,phiQa_v2]


    #Compression of R1

    e = (p**2 - 1)//D5

    criteria_1 = tate_pairing_pari(P5,Q5,D5)** e
    
    phiRce1 = tate_pairing_pari(S1[1][2],Q5,D5)** e
    phiRce2 = tate_pairing_pari(P5,S1[1][2],D5)** e

    phiRc_v1 = discrete_log_pari(phiRce1,criteria_1,D5)//(5**c_ex)
    phiRc_v2 = discrete_log_pari(phiRce2,criteria_1,D5)//(5**c_ex)

    if phiRc_v1 % 5 == 0:

        inv_phiR = inverse_mod(phiRc_v2,5**c)

        phiRc_v1 = (phiRc_v1 * inv_phiR) % 5**c
        
        phiR1 = [0,phiRc_v1]

    else:

        inv_phiR = inverse_mod(phiRc_v1,5**c)
        
        phiRc_v2 = (phiRc_v2 * inv_phiR) % 5**c
        
        phiR1 = [1,phiRc_v2]
    

    #Compression of P2 Q2
    e = (p**2 - 1)//D2
    
    S2 = MontoEllip(cipher[1][0],[cipher[1][1],cipher[1][2],cipher[1][3]])

    #P, Q = entangled_torsion_basis(S2[0], elligator_tables, (p+1)//(D2*D5))
    P, Q = torsion_basis(S2[0], D2*D5)

    P2 = D5 * P
    Q2 = D5 * Q
    P5 = D2 * P
    Q5 = D2 * Q

    criteria_2 = tate_pairing_pari(P2,Q2,D2)**e
    
    phiPae1 = tate_pairing_pari(S2[1][0],Q2,D2)**e
    phiPae2 = tate_pairing_pari(P2,S2[1][0],D2)**e
    phiQae1 = tate_pairing_pari(S2[1][1],Q2,D2)**e
    phiQae2 = tate_pairing_pari(P2,S2[1][1],D2)**e

    phiPa_v1_ = windowed_pohlig_hellman(phiPae1,criteria_2,a+2+a_ex,window)//(2**a_ex)
    phiPa_v2_ = windowed_pohlig_hellman(phiPae2,criteria_2,a+2+a_ex,window)//(2**a_ex)
    phiQa_v1_ = windowed_pohlig_hellman(phiQae1,criteria_2,a+2+a_ex,window)//(2**a_ex)
    phiQa_v2_ = windowed_pohlig_hellman(phiQae2,criteria_2,a+2+a_ex,window)//(2**a_ex)


    if phiPa_v1_ % 2 == 0:

        inv_phi = inverse_mod(phiPa_v2_,2**(a+2))
        
        phiPa_v1_ = (phiPa_v1_ * inv_phi) % 2**(a+2)
        phiQa_v1_ = (phiQa_v1_ * inv_phi) % 2**(a+2)
        phiQa_v2_ = (phiQa_v2_ * inv_phi) % 2**(a+2)
        
        phiA2 = [0,phiPa_v1_,phiQa_v1_,phiQa_v2_]

    else:

        inv_phi = inverse_mod(phiPa_v1_,2**(a+2))
        
        phiPa_v2_ = (phiPa_v2_ * inv_phi) % 2**(a+2)
        phiQa_v1_ = (phiQa_v1_ * inv_phi) % 2**(a+2)
        phiQa_v2_ = (phiQa_v2_ * inv_phi) % 2**(a+2)
        
        phiA2 = [1,phiPa_v2_,phiQa_v1_,phiQa_v2_]


    #Compression of R2

    e = (p**2 - 1)//D5

    criteria_2 = tate_pairing_pari(P5,Q5,D5)** e
    
    phiRce1_ = tate_pairing_pari(S2[1][2],Q5,D5)** e
    phiRce2_ = tate_pairing_pari(P5,S2[1][2],D5)** e

    phiRc_v1_ = discrete_log_pari(phiRce1_,criteria_2,D5)//(5**c_ex)
    phiRc_v2_ = discrete_log_pari(phiRce2_,criteria_2,D5)//(5**c_ex)

    phiRc_v1_ = (phiRc_v1_ * inv_phiR) % 5**c
    phiRc_v2_ = (phiRc_v2_ * inv_phiR) % 5**c

    phiR2 = [phiRc_v1_,phiRc_v2_]

    return [[cipher[0][0],phiA1,phiR1],[cipher[1][0],phiA2,phiR2]]


def decomp_cipher(cipher_comp,Alice_secret,a,c,p,Fp,Fp2):

    S1 = MontoEllip(cipher_comp[0][0],[])

    a_ex = 0
    check_ex = (p+1)//(2**(a+2))
    while check_ex % 2 == 0:
        a_ex += 1
        check_ex = check_ex // 2

    c_ex = 0
    check_ex = (p+1)//(5**c)
    while check_ex % 5 == 0:
        c_ex += 1
        check_ex = check_ex // 5

    D2 = 2**(a+2+a_ex)
    D5 = 5**(c+c_ex)
    
    #P, Q = entangled_torsion_basis(S1[0], elligator_tables, (p+1)//(D2*D5))
    P, Q = torsion_basis(S1[0], D2*D5)
    P1 = (2**a_ex) * D5 * P
    Q1 = (2**a_ex) * D5 * Q
    P5 = (5**c_ex) * D2 * P
    Q5 = (5**c_ex) * D2 * Q

    #Decompression of P1 Q1
    if cipher_comp[0][1][0] == 0:

        Pa1_point = (cipher_comp[0][1][1]) * P1 + Q1
        Qa1_point = (cipher_comp[0][1][2]) * P1 + (cipher_comp[0][1][3]) * Q1

    else:

        Pa1_point = P1 + (cipher_comp[0][1][1]) * Q1
        Qa1_point = (cipher_comp[0][1][2]) * P1 + (cipher_comp[0][1][3]) * Q1

    #Decompression of R1
    if cipher_comp[0][2][0] == 0:

        Rc1_point = (cipher_comp[0][2][1]) * P5 + Q5

    else:

        Rc1_point = P5 + (cipher_comp[0][2][1]) * Q5


    S2 = MontoEllip(cipher_comp[1][0],[])

    #P, Q = entangled_torsion_basis(S2[0], elligator_tables, (p+1)//(D2*D5))
    P, Q = torsion_basis(S2[0], D2*D5)
    P2 = (2**a_ex) * D5 * P
    Q2 = (2**a_ex) * D5 * Q
    P5 = (5**c_ex) * D2 * P
    Q5 = (5**c_ex) * D2 * Q

    #Decompression of P2 Q2
    if cipher_comp[1][1][0] == 0:

        Pa2_point = (cipher_comp[1][1][1]) * P2 + Q2
        Qa2_point = (cipher_comp[1][1][2]) * P2 + (cipher_comp[1][1][3]) * Q2

    else:

        Pa2_point = P2 + (cipher_comp[1][1][1]) * Q2
        Qa2_point = (cipher_comp[1][1][2]) * P2 + (cipher_comp[1][1][3]) * Q2


    Pa1_point = Alice_secret[0] * Pa1_point
    Qa1_point = Alice_secret[1] * Qa1_point

    value1 = (weil_pairing_pari(Pa1_point,Qa1_point,2**(a+2)))**(2**(2*a)-9)
    value2 = (weil_pairing_pari(Pa2_point,Qa2_point,2**(a+2)))
    
    window = [c//10]
    a02 = windowed_pohlig_hellman(value1,value2,a+2,window)

    if not mod(a02,2**(a+2)).is_square():
        a02 = (-1)*a02

    a0 = mod(a02,2**(a+2)).sqrt()

    Pa2_point = (Alice_secret[0] * Integer(a0)) * Pa2_point
    Qa2_point = (Alice_secret[1] * Integer(a0)) * Qa2_point

    #Decompression of R2
    Rc2_point = (cipher_comp[1][2][0]) * P5 + (cipher_comp[1][2][1]) * Q5

    Pa1 = [Pa1_point[0],Pa1_point[2]]
    Qa1 = [Qa1_point[0],Qa1_point[2]]
    Pa2 = [Pa2_point[0],Pa2_point[2]]
    Qa2 = [Qa2_point[0],Qa2_point[2]]
    Rc1 = [Rc1_point[0],Rc1_point[2]]
    Rc2 = [Rc2_point[0],Rc2_point[2]]
    

    return [[cipher_comp[0][0],Pa1,Qa1,Rc1],[cipher_comp[1][0],Pa2,Qa2,Rc2]]
