from random import randint
from math import floor
from cysignals.alarm import alarm, AlarmInterrupt, cancel_alarm 
from sage.all import EllipticCurve
from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
from cornacchia.cornacchia import *
from montgomery import *
from theta_isogeny import *


#Sometimes the primary test in the the Cornacchia alogrithm takes too much time,
#so we use a timeout.
def generate_gamma(a,b,tau,p):

    N = tau*(3**b)

    m = floor((N/(p*2))**(1/2))

    List = []

    while (len(List) == 0):
        z = 0
        t = 0
        while (z%3 == 0) and (t%3 == 0):
            z = randint(-m,m)
            t = randint(-m,m)
        NN = N - p*(z**2+t**2)
        try:
            alarm(3)
            List = Cornacchia(NN,1)
        except AlarmInterrupt:
            z = 0
            t = 0
        cancel_alarm()

    return [List[0],List[1],z,t]


#PQa = [A,Pa,Qa]
#PQa1 = [A1,Pa1,Qa1]
#PQa_shift = [Pa_shift,Qa_shift,Pa1_shift,Qa1_shift]
def compute_image_gamma0(PQa,PQa1,PQa_shift,given_points,given_points_shift,a,tau,sqrtminus,strategy_for_2dim):

    if len(PQa_shift) != 4:
        return None

    if len(given_points) != len(given_points_shift):
        return None

    A = PQa[0]
    Pa = PQa[1]
    Qa = PQa[2]
    Pa = scalar(Pa,A,3)
    Qa = scalar(Qa,A,3)
    Pa_shift = scalar(PQa_shift[0],A,3)
    Qa_shift = scalar(PQa_shift[1],A,3)

    A1 = PQa1[0]
    Pa1 = PQa1[1]
    Qa1 = PQa1[2]
    Pa_shift1 = PQa_shift[2]
    Qa_shift1 = PQa_shift[3]


    fourtorsion = scalar(Qa,A,2**a)
    fourtorsion1 = scalar(Qa1,A1,2**a)

    #[eight_torsion,generator_of_kernel,shift]
    points_for_kernel_A = [scalar(Pa,A,2**(a-1)),scalar(Qa,A,2**(a-1)),Pa,Qa,Pa_shift,Qa_shift]
    points_for_kernel_A1 = [scalar(Pa1,A1,2**(a-1)),scalar(Qa1,A1,2**(a-1)),Pa1,Qa1,Pa_shift1,Qa_shift1]

    zero_points = []
    for i in range(len(given_points)):
        zero_points.append([1,0])

    shift_points = []
    for i in range(len(given_points)):
        shift_points.append([1,1])

    allpoints_A = points_for_kernel_A + given_points + given_points_shift
    allpoints_A1 = points_for_kernel_A1 + zero_points + shift_points
    
    theta1,points = ProdtoTheta(A,A1,fourtorsion,fourtorsion1,allpoints_A,allpoints_A1)

    eight_torsion1 = points[:2]
    kernel1 = points[2:4]
    kernel_shift1 = points[4:6]

    given_points_theta1 = points[6:6+len(given_points)]
    given_points_theta_shift1 = points[6+len(given_points):6+len(given_points)*2]


    gammapoints_theta = hatPhiPsi_scalar(given_points_theta1,given_points_theta_shift1,eight_torsion1,kernel1,kernel_shift1,theta1,a,strategy_for_2dim)

    AA_AA1 = ThetatoMonProd(theta1,gammapoints_theta+[eight_torsion1[0]],sqrtminus)

    points_result = []
    for i in range(len(given_points)):
        points_result.append(AA_AA1[1][i])
    
    return points_result
