from random import randint
from sage.all import EllipticCurve
from sage.all import discrete_log
from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius
from theta_isogeny import *
from compute_gamma import *
from utilities.pairing import weil_pairing_pari
import time


def set_param(lam):

    a = lam * 3

    if lam == 128:
        b = 162
        c = 56
        f = 30
    elif lam == 192:
        b = 243
        c = 83
        f = 118
    elif lam == 256:
        b = 324
        c = 111
        f = 436
    else:
        raise ValueError("lam should be 128 or 192 or 256.")

    return a,b,c,f


def generate_pubparam(a,b,c,sqrtminus,Fp,Fp2,p):

    #The curve of j-invariant 1728
    A = [Fp2(1),2]
    
    checker = 1
    while checker == 1:
        xP = Fp.random_element()
        P = [xP,1]
        P = scalar(P,A,(p+1)//((2**(a+2))*(3**b)*(5**c)))
        Pa = scalar(P,A,((2**(a+1))*(3**b)*(5**c)))
        Pb = scalar(P,A,((2**(a+2))*(3**(b-1))*(5**c)))
        Pc = scalar(P,A,((2**(a+2))*(3**b)*(5**(c-1))))
        if (Pa[1] != 0) and (Pb[1] != 0) and (Pc[1] != 0):
            checker = 0

    Pa = scalar(P,A,(3**b)*(5**c))
    Pbc = scalar(P,A,2**(a+2))
    Qbc = [Pbc[0]*(-1),Pbc[1]]

    #[2**a]Pa = (1,*).
    Pa_ = scalar(Pa,A,2**a)
    if Pa_[0] != Pa_[1]:
        Pa[0] *= -1

    checker = 1
    while checker == 1:
        xP = Fp2.random_element()
        Qa = [xP,1]
        Qa = scalar(Qa,A,(p+1)//(2**(a+2)))
        Q = scalar(Qa,A,2**(a+1))
        if (Q[0]**2 == (-1)*Q[1]**2) and (Q[1] != 0):
            checker = 0

    return [Pbc,Qbc,Pa,Qa]


def generate_pubkey(a,b,c,pub_param,strategy,strategy_for_2dim,sqrtminus,Fp,Fp2,p):

    tau = 2**(2*a) - 9
    gamma = generate_gamma(a,b,tau,p)

    Pbc = pub_param[0]
    Qbc = pub_param[1]
    Pa = pub_param[2]
    Qa = pub_param[3]

    A = [Fp2(1),2]

    Pb = scalar(Pbc,A,5**c)
    Qb = scalar(Qbc,A,5**c)
    Pc = scalar(Pbc,A,3**b)
    Qc = scalar(Qbc,A,3**b)

    E,S = MontoEllip(A,[Pb,Qb,Pa,Qa,Pc,Qc])

    a1 = randint(0,2**(a+2)-1)
    a2 = randint(0,2**(a+1)-1)*2 + 1
    S[3] = a1*S[2] + a2*S[3]
    Qa = [S[3][0],S[3][2]]

    c1 = 0
    c2 = 0
    while (c1%5 == 0) and (c2%5 == 0):
        c1 = randint(0,5**c-1)
        c2 = randint(0,5**c-1)
    S[4] = c1 * S[4] + c2 * S[5]
    R = [S[4][0],S[4][2]]

    pi = EllipticCurveHom_frobenius(E)

    iota = WeierstrassIsomorphism(E, (sqrtminus,0,0,0))

    gammaPb = gamma[0]*S[0]+gamma[1]*iota(S[0])+gamma[2]*pi(S[0])+gamma[3]*pi(iota(S[0]))
    gammaQb = gamma[0]*S[1]+gamma[1]*iota(S[1])+gamma[2]*pi(S[1])+gamma[3]*pi(iota(S[1]))
    gammaPa = gamma[0]*S[2]+gamma[1]*iota(S[2])+gamma[2]*pi(S[2])+gamma[3]*pi(iota(S[2]))
    gammaQa = gamma[0]*S[3]+gamma[1]*iota(S[3])+gamma[2]*pi(S[3])+gamma[3]*pi(iota(S[3]))
    gammaR = gamma[0]*S[4]+gamma[1]*iota(S[4])+gamma[2]*pi(S[4])+gamma[3]*pi(iota(S[4]))

    if gammaQb.order() < gammaPb.order():
        #s = gammaPb.discrete_log(gammaQb,3**b)
        s = discrete_log(gammaQb,gammaPb,3**b,operation='+')
        kernel1 = S[1] - s*S[0]
    else:
        #s = gammaQb.discrete_log(gammaPb,3**b)
        s = discrete_log(gammaPb,gammaQb,3**b,operation='+')
        kernel1 = S[0] - s*S[1]
    kernel1x = [kernel1[0],kernel1[2]]

    #Points for constructing the isogeny diamond

    A1 = [Fp2(1),2]
    Pa1 = [gammaPa[0],gammaPa[2]]
    Qa1 = [gammaQa[0],gammaQa[2]]
    R1 = [gammaR[0],gammaR[2]]

    #Set [2^a]Pa1 = (1,*)
    Pa1_ord4 = scalar(Pa1,A1,2**a)
    if Pa1_ord4[0] != Pa1_ord4[1]:
        Pa1[0] = (-1)*Pa1[0]
        Qa1[0] = (-1)*Qa1[0]
        R1[0] = (-1)*R1[0]

    AP1 = isog_3_strategy(b,A1,[Pa,Qa,R],kernel1x,strategy)
    A = AP1[0]
    Pa = AP1[1][0]
    Qa = AP1[1][1]
    R = AP1[1][2]

    #We obtained (A,Pa,Qa,R) and (A1,Pa1,Qa1,R1).

    PQa,PQa1,PQa_shift,PQball,PQb_shift = get_PQb_and_shift(A,A1,Pa,Qa,Pa1,Qa1,tau,Fp2,a,b,p)

    Pb = PQball[0]
    Qb = PQball[1]
    PQb = PQball[2]

    pointsPQb = compute_image_gamma0(PQa,PQa1,PQa_shift,PQball,PQb_shift,a,tau,sqrtminus,strategy_for_2dim)

    Pb1 = pointsPQb[0]
    Qb1 = pointsPQb[1]
    PQb1 = pointsPQb[2]
    Qback = [Fp2(0),1]

    for i in range(6):

        backtrack_check = 0
        while (backtrack_check == 0):
            s = 3*randint(0,3**(b-1)-1) + randint(1,2)

            kernelx = Ladder_3pt(Pb,Qb,PQb,A,s)

            kernel_check = scalar(kernelx,A,3**(b-1))

            if kernel_check[0]/kernel_check[1] != Qback[0]/Qback[1]:
                
                backtrack_check = 1

        kernel1x = Ladder_3pt(Pb1,Qb1,PQb1,A1,s)
        
        AP = isog_3_strategy(b,A,[Pa,Qa,R,Qb],kernelx,strategy)
        A = AP[0]
        Pa = AP[1][0]
        Qa = AP[1][1]
        R = AP[1][2]
        Qback = AP[1][3]
        Qback = scalar(Qback,A,3**(b-1))

        AP1 = isog_3_strategy(b,A1,[Pa1,Qa1,R1],kernel1x,strategy)
        A1 = AP1[0]
        Pa1 = AP1[1][0]
        Qa1 = AP1[1][1]
        R1 = AP1[1][2]

        PQa,PQa1,PQa_shift,PQball,PQb_shift = get_PQb_and_shift(A,A1,Pa,Qa,Pa1,Qa1,tau,Fp2,a,b,p)

        Pb = PQball[0]
        Qb = PQball[1]
        PQb = PQball[2]

        pointsPQb = compute_image_gamma0(PQa,PQa1,PQa_shift,PQball,PQb_shift,a,tau,sqrtminus,strategy_for_2dim)

        Pb1 = pointsPQb[0]
        Qb1 = pointsPQb[1]
        PQb1 = pointsPQb[2]


    alpha = 5 * randint(0,5**(c-1) - 1) + randint(1,4)
    R1 = scalar(R1,A1,alpha)

    Alice_secret = [2*randint(0,2**(a+2)-1)+1,2*randint(0,2**(a+2)-1)+1,alpha]
    Pa1 = scalar(Pa1,A1,Alice_secret[0])
    Qa1 = scalar(Qa1,A1,Alice_secret[1])

    pubkey = [[Pa,Qa,Pb,Qb,PQb,R],[Pa1,Qa1,Pb1,Qb1,PQb1,R1]]
    
    return pubkey,Alice_secret


def encrypt(a,b,c,strategy,sqrtminus,Fp,Fp2,pubkey,mu):

    Pa = pubkey[0][0]
    Qa = pubkey[0][1]
    Pb = pubkey[0][2]
    Qb = pubkey[0][3]
    PQb = pubkey[0][4]
    R = pubkey[0][5]
    A = get_A(Pb,Qb,PQb)

    Pa1 = pubkey[1][0]
    Qa1 = pubkey[1][1]
    Pb1 = pubkey[1][2]
    Qb1 = pubkey[1][3]
    PQb1 = pubkey[1][4]
    R1 = pubkey[1][5]
    A1 = get_A(Pb1,Qb1,PQb1)

    s = 3*randint(0,3**(b-1)-1) + randint(1,2)

    kernelx = Ladder_3pt(Pb,Qb,PQb,A,s)
    kernel1x = Ladder_3pt(Pb1,Qb1,PQb1,A1,s)

    AP = isog_3_strategy(b,A,[Pa,Qa,R],kernelx,strategy)
    A = AP[0]
    Pa = AP[1][0]
    Qa = AP[1][1]
    R = AP[1][2]

    AP1 = isog_3_strategy(b,A1,[Pa1,Qa1,R1],kernel1x,strategy)
    A1 = AP1[0]
    Pa1 = AP1[1][0]
    Qa1 = AP1[1][1]
    R1 = AP1[1][2]

    beta = 5 * randint(0,5**(c-1) - 1) + randint(1,4)
    R = scalar(R,A,beta)
    R1 = scalar(R1,A1,beta*mu)

    Bob_secret = [2*randint(0,2**(a+2)-1)+1,2*randint(0,2**(a+2)-1)+1]
    Pa = scalar(Pa,A,Bob_secret[0])
    Qa = scalar(Qa,A,Bob_secret[1])
    Pa1 = scalar(Pa1,A1,Bob_secret[0])
    Qa1 = scalar(Qa1,A1,Bob_secret[1])

    return [[A,Pa,Qa,R],[A1,Pa1,Qa1,R1]]


def decrypt(a,b,c,strategy_for_2dim,sqrtminus,Fp,Fp2,cipher,Alice_secret):

    tau = 2**(2*a) - 9
    
    A = cipher[0][0]
    Pa = cipher[0][1]
    Qa = cipher[0][2]
    R = cipher[0][3]

    alpha = Alice_secret[2]
    R = scalar(R,A,alpha)

    Pa = scalar(Pa,A,Alice_secret[0])
    Qa = scalar(Qa,A,Alice_secret[1])

    A1 = cipher[1][0]
    Pa1 = cipher[1][1]
    Qa1 = cipher[1][2]
    R1 = cipher[1][3]

    PQa = [A,Pa,Qa]
    PQa1 = [A1,Pa1,Qa1]

    E,S = MontoEllip(A,[Pa,Qa,R])
    E1,S1 = MontoEllip(A1,[Pa1,Qa1])

    T1_0 = (2**a)*S[0]
    T1_1 = (2**a)*S1[0]
    if (weil_pairing_pari(T1_0,(2**a)*S[1],4)** 3 != weil_pairing_pari(T1_1,(2**a)*S1[1],4)):
        S[0] = (-1)*S[0]
        T1_0 = (-1)*T1_0

    Rs = S[2] + T1_0
    R_shift = [Rs[0],Rs[2]]
    Pas = S[0] + T1_0
    Qas = S[1] + T1_0
    Pas1 = S1[0] + T1_1
    Qas1 = S1[1] + T1_1
    
    Pa_shift = [Pas[0],Pas[2]]
    Pa1_shift = [Pas1[0],Pas1[2]]
    Qa_shift = [Qas[0],Qas[2]]
    Qa1_shift = [Qas1[0],Qas1[2]]
    PQa_shift = [Pa_shift,Qa_shift,Pa1_shift,Qa1_shift]

    R_ = [R]
    R_shift_ = [R_shift]

    pointsR = compute_image_gamma0(PQa,PQa1,PQa_shift,R_,R_shift_,a,tau,sqrtminus,strategy_for_2dim)

    R1_2 = pointsR[0]
    E1,S2 = MontoEllip(A1,[R1,R1_2])
    #mu = S2[1].discrete_log(S2[0],5**c)
    mu = discrete_log(S2[0],S2[1],5**c,operation='+')

    return mu
