from sage.all import EllipticCurve
from montgomery import *


def S_operator(List):

    List2 = List.copy()
    for i in range(len(List)):
        List2[i] = List2[i]**2

    return List2


def H_operator(List):

    if len(List) != 4:
        return None

    List2 = [0,0,0,0]

    List2[0] = sum(List)
    List2[1] = List[0] - List[1] + List[2] - List[3]
    List2[2] = List[0] + List[1] - List[2] - List[3]
    List2[3] = List[0] - List[1] - List[2] + List[3]

    return List2

def elementwisemult(List,List2):

    List_result = List.copy()
    if len(List) != len(List2):
        return None

    for i in range(len(List)):
        List_result[i] *= List2[i]

    return List_result

def List_tensor(List1,List2):

    List = []
    for i in range(len(List1)):
        for j in range(len(List2)):
            List.append(List1[i]*List2[j])
    
    return List

def Action_matrix(matrix,List):

    List_result = []
    for i in range(len(matrix)):
        List2 = elementwisemult(matrix[i],List)
        element = 0
        for j in range(len(List2)):
            element += List2[j]

        List_result.append(element)

    return List_result


def proj_inv(theta):

    if len(theta) == 1:
        return [1]
    
    if len(theta) == 2:
        return [theta[1],theta[0]]

    if len(theta) == 3:
        return [theta[1]*theta[2],theta[2]*theta[0],theta[0]*theta[1]]

    m = len(theta)//2
    result1 = proj_inv(theta[:m])
    result2 = proj_inv(theta[m:])
    
    left = result1[0]*theta[0]
    right = result2[-1]*theta[-1]

    for i in range(len(result1)):
        result1[i] *= right
    
    for i in range(len(result2)):
        result2[i] *= left

    return result1+result2


#Compute [a0^{-1},...,an^{-1}] from [a0,...,an] 
def batched_inv(List):

    mult_list = [List[0]]
    for i in range(1,len(List)):
        mult_list.append(mult_list[-1] * List[i])

    lastelement = 1/mult_list[-1]

    inverse_list = [lastelement]
    
    for i in range(1,len(List)):
        inverse_list.append(inverse_list[-1]*List[len(List)-i])

    inverse_list.reverse()
    
    for i in range(1,len(List)):
        inverse_list[i] *= mult_list[i-1]
    
    return inverse_list


#We take ((1:1),(1:1)) and (four_torsion1,four_torsion2) to determine the theta structure of E1  * E2
def ProdtoTheta(A24_1,A24_2,four_torsion1,four_torsion2,points1,points2):

    two_torsion1 = double(four_torsion1,A24_1)
    two_torsion2 = double(four_torsion2,A24_2)

    WXWZUXUZ1 = List_tensor(two_torsion1[::-1],four_torsion1)
    WXWZUXUZ2 = List_tensor(two_torsion2[::-1],four_torsion2)
    delta1 = WXWZUXUZ1[0] - WXWZUXUZ1[3]
    delta2 = WXWZUXUZ2[0] - WXWZUXUZ2[3]

    List = [delta1,four_torsion1[1],delta2,four_torsion2[1]]
    List = batched_inv(List)

    action1 = [-1*WXWZUXUZ1[3]*List[0],-1*WXWZUXUZ1[1]*List[0],WXWZUXUZ1[2]*List[0] - four_torsion1[0]*List[1]]
    action1.append(-1*action1[0])
    action2 = [-1*WXWZUXUZ2[3]*List[2],-1*WXWZUXUZ2[1]*List[2],WXWZUXUZ2[2]*List[2] - four_torsion2[0]*List[3]]
    action2.append(-1*action2[0])
    
    matrix0_ = [action1[0]*action2[0] + action1[2]*action2[2] + 1, action1[0]*action2[2] + action1[2]*action2[0]]
    matrix1_tensor = List_tensor(matrix0_,action2)
    matrix0 = [matrix0_[0],matrix0_[1],matrix0_[1],matrix0_[0]]
    matrix1 = [matrix1_tensor[0] + matrix1_tensor[5], matrix1_tensor[2] + matrix1_tensor[7], matrix1_tensor[4] + matrix1_tensor[1], matrix1_tensor[6] + matrix1_tensor[3]]
    matrix2 = [matrix0[2],matrix0[3],matrix0[0],matrix0[1]]
    matrix3 = [matrix1[2],matrix1[3],matrix1[0],matrix1[1]]
    
    matrix = [matrix0,matrix1,matrix2,matrix3]
    theta = [matrix[0][0],matrix[1][0],matrix[2][0],matrix[3][0]]

    points = []
    if len(points1) == len(points2):
        for i in range(len(points1)):
            points.append(List_tensor(points1[i],points2[i]))
            points[i] = Action_matrix(matrix,points[i])


    return theta,points


def MontoTheta(four_torsion,points):

    #theta-null
    theta = [four_torsion[0]+four_torsion[1],four_torsion[0]-four_torsion[1]]

    for i in range(len(points)):
    
        pointi = [0,0]
        pointi[0] = theta[0] * (points[i][0]-points[i][1])
        pointi[1] = theta[1] * (points[i][0]+points[i][1])

        point[i] = pointi.copy()
    
    return theta,points

#[a,-b] is [0,1]
def ThetatoMon1(theta,points):

    points_result = []

    for i in range(len(points)):
    
        pointi = [0,0]
        pointi[0] = theta[1] * points[i][0] + theta[0] * points[i][1]
        pointi[1] = theta[1] * points[i][0] - theta[0] * points[i][1]

        points_result.append(pointi.copy())
    
    return points_result

#[b,-a] is [0,1]
def ThetatoMon2(theta,points,sqrtminus):

    points_result = []

    for i in range(len(points)):

        pointi = [0,0]
        pointi[0] = (theta[0] * points[i][0] + theta[1] * points[i][1])*(-1)*sqrtminus
        pointi[1] = theta[1] * points[i][0] - theta[0] * points[i][1]

        points_result.append(pointi.copy())
    
    return points_result

#We assume delta = 0
def ThetatoMonProd(theta,points,sqrtminus):

    matrix = [[1,0,0,0],[0,sqrtminus,0,0],[0,0,sqrtminus,0],[0,0,0,1]]
    theta2 = H_operator(theta)
    theta2 = Action_matrix(matrix,theta2)

    theta_A1 = [theta2[1],theta2[3]]
    theta_A2 = [theta2[0],theta2[1]]

    points1_theta = []
    points2_theta = []
    for i in range(len(points)):

        pointi = H_operator(points[i])
        pointi = Action_matrix(matrix,pointi)
        points1_theta.append([pointi[1],pointi[3]])
        points2_theta.append([pointi[0],pointi[1]])

    points1 = ThetatoMon1(theta_A1,points1_theta)
    points2 = ThetatoMon2(theta_A2,points2_theta,sqrtminus)
    
    return [points1,points2]


def Prodtheta(theta1,theta2,points1,points2):

    theta = List_tensor(theta1,theta2)
    if len(points1) != len(points2):
        return theta,None

    point = []
    for i in range(len(points1)):
        point.append(List_tensor(points1[i],points2[i]))
    
    return theta,point

#6M+8S+**
def theta_doubling(theta_inv,theta_dual_invsq,point):

    point_result = point.copy()
    if (len(point) != 4) or (len(theta_dual_invsq) != 4) or (len(theta_inv) != 4):
        return None

    point_result = H_operator(S_operator(point_result))
    point_result = S_operator(point_result)

    point_result = elementwisemult(point_result,theta_dual_invsq)
    point_result = H_operator(point_result)
    point_result = elementwisemult(point_result,theta_inv)
    
    return point_result

def theta_doublings(theta,theta_dual_invsq,point,a):

    point_result = point.copy()
    
    theta_inv = [1]
    theta_inv += batched_inv(theta[1:4])
    for i in range(1,4):
        theta_inv[i] *= theta[0]

    for i in range(a):
        point_result = theta_doubling(theta_inv,theta_dual_invsq,point_result)
    
    return point_result


def EvalIsogeny(theta_dual_inv,points):

    points_result = points.copy()
    
    for i in range(len(points)):
        points_result[i] = H_operator(S_operator(points_result[i]))
        points_result[i] = elementwisemult(theta_dual_inv,points_result[i])
        points_result[i] = H_operator(points_result[i])

    return points_result

#We only have the case of delta = 0.
def EvalIsogeny_special(theta_dual_inv,points,points_shift):

    points_result = points.copy()
    if len(points) != len(points_shift):
        return None

    for i in range(len(points)):
        points_result[i] = H_operator(S_operator(points_result[i]))
        points_shift[i] = H_operator(S_operator(points_shift[i]))
        points_result[i] = elementwisemult(theta_dual_inv,points_result[i])

        if points_result[i][1] != 0:
            lambda_inv = points_result[i][1] / (theta_dual_inv[0] * points_shift[i][0])
        else:
            lambda_inv = points_result[i][0] / (theta_dual_inv[1] * points_shift[i][1])

        points_result[i][3] = lambda_inv * theta_dual_inv[2] * points_shift[i][2]
        points_result[i] = H_operator(points_result[i])

    return points_result


#29M+8S+1I
#We possibly have the case of delta = 0.
def CodomainIsogeny(eight_torsions):

    eight_torsions_sub = [[0,0,0,0],[0,0,0,0]]
    eight_torsions_sub[0] = H_operator(S_operator(eight_torsions[0]))
    eight_torsions_sub[1] = H_operator(S_operator(eight_torsions[1]))

    if eight_torsions_sub[0][3] == 0:

        alpha0 = eight_torsions_sub[1][0]
        gamma0 = eight_torsions_sub[0][0]

        alpha = eight_torsions_sub[0][0] * alpha0
        beta = eight_torsions_sub[0][1] * alpha0
        gamma = eight_torsions_sub[1][2] * gamma0
        
        theta_dual = [alpha,beta,gamma,0]

        theta_dual_inv = proj_inv([alpha,beta,gamma])+[0]
        
    else:
        invList = proj_inv([eight_torsions_sub[0][0],eight_torsions_sub[1][0],eight_torsions_sub[1][1]])
        
        alpha = eight_torsions_sub[0][0] * invList[0]
        beta = eight_torsions_sub[0][1] * invList[0]
        gamma = eight_torsions_sub[1][2] * invList[1] * alpha
        delta = eight_torsions_sub[1][3] * invList[2] * beta
        beta *= alpha
        alpha = alpha**2
        theta_dual = [alpha,beta,gamma,delta]

        theta_dual_inv = proj_inv(theta_dual)

    theta = H_operator(theta_dual)

    return theta_dual,theta_dual_inv,theta


#compute a (2^a,2^a)-isogeny from a non-product of elliptic curves.
def isogchain_strategy(theta_dual_inv,theta,kernelpoints,a,strategy,mode):
    
    S = [[a,kernelpoints.copy()]]
    i = 0
    isogeny_chain = []
    isogeny_chain_dual = []

    theta_changed = 1

    while len(S) != 0:

        hP = S.pop()
        h = hP[0]
        kernelpoints = hP[1].copy()

        if h == 1:
            theta_dual,theta_dual_inv,theta = CodomainIsogeny(kernelpoints)
            S2 = []

            theta_changed = 1

            if mode == 0:
                isogeny_chain.append(theta_dual_inv.copy())
            elif mode == 1:
                isogeny_chain.append(proj_inv(theta))
            else:
                isogeny_chain.append(theta_dual_inv.copy())
                isogeny_chain_dual.append(proj_inv(theta))
            
            while len(S) != 0:
                hP = S.pop(0)
                kernelpoints = EvalIsogeny(theta_dual_inv,hP[1])
                S2.append([hP[0]-1,kernelpoints.copy()])

            S = S2.copy()

        elif (strategy[i] < h) and (0 < strategy[i]):

            if theta_changed == 1:
                #theta_dual_invsq = [1]
                theta_dual_invsq_sub = H_operator(S_operator(theta))
                #theta_dual_invsq += batched_inv(theta_dual_invsq_sub[1:])
                #for j in range(1,4):
                #    theta_dual_invsq[j] *= theta_dual_invsq_sub[0]
                theta_dual_invsq = proj_inv(theta_dual_invsq_sub)

            theta_changed = 0

            S.append([h,kernelpoints.copy()])
            
            kernelpoints[0] = theta_doublings(theta,theta_dual_invsq,kernelpoints[0],strategy[i])
            kernelpoints[1] = theta_doublings(theta,theta_dual_invsq,kernelpoints[1],strategy[i])
            S.append([h-strategy[i],kernelpoints.copy()])
            i += 1

    if (mode != 0) and (mode != 1):
        return isogeny_chain,isogeny_chain_dual
    
    return isogeny_chain


#Compute \hat{\Phi} \circ \Psi(P)
#We compute the isogeny by using 2^(a+2)-torsion points.
def hatPhiPsi_scalar(points,points_shift,eight_torsion1,kernel1,kernel_shift1,theta1,a,strategy):
    
    #Compute Psi

    isogeny_chain_dual = [proj_inv(theta1)]

    theta_dual,theta_dual_inv,theta2 = CodomainIsogeny(eight_torsion1)

    isogeny_chain_dual.append(proj_inv(theta2))

    isogeny_chain = [theta_dual_inv.copy()]
    
    kernel1 = EvalIsogeny_special(theta_dual_inv,kernel1,kernel_shift1)
    
    isogeny_chain1,isogeny_chain_dual1 = isogchain_strategy(theta_dual_inv,theta2,kernel1,a-1,strategy,2)

    isogeny_chain += isogeny_chain1

    isogeny_chain_dual += isogeny_chain_dual1

    #test = proj_inv(isogeny_chain[-1])
    #test2 = H_operator(test)
    #print(test[0]/test[3] == test2[0]/test2[3])
    
    matrix = [[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]]

    #Evaluate points
    points = EvalIsogeny_special(isogeny_chain[0],points,points_shift)
    
    for i in range(1,len(isogeny_chain)):
        points =  EvalIsogeny(isogeny_chain[i],points)

    for i in range(len(points)):
        points[i] = Action_matrix(matrix,points[i])

    for i in range(1,len(isogeny_chain_dual)):
        points =  EvalIsogeny(isogeny_chain_dual[-i-1],points)

    for i in range(len(points)):
        points[i] = H_operator(points[i])

    return points
